@@ -109,6 +109,9 @@ var AST_Node = DEFNODE("Node", "start end", {
         start: "[AST_Token] The first token of this node",
         end: "[AST_Token] The last token of this node"
     },
+    equals: function(node) {
+        return this.TYPE == node.TYPE && this._equals(node);
+    },
     walk: function(visitor) {
         visitor.visit(this);
     },
@@ -138,6 +141,7 @@ var AST_Node = DEFNODE("Node", "start end", {
 }, null);
 
 DEF_BITPROPS(AST_Node, [
+    // AST_Node
     "_optimized",
     "_squeezed",
     // AST_Call
@@ -172,6 +176,8 @@ DEF_BITPROPS(AST_Node, [
     "pure",
     // AST_Assign
     "redundant",
+    // AST_Node
+    "single_use",
     // AST_ClassProperty
     "static",
     // AST_Call
@@ -231,6 +237,24 @@ AST_Node.disable_validation = function() {
     while (restore = restore_transforms.pop()) restore();
 };
 
+function all_equals(k, l) {
+    return k.length == l.length && all(k, function(m, i) {
+        return m.equals(l[i]);
+    });
+}
+
+function list_equals(s, t) {
+    return s.length == t.length && all(s, function(u, i) {
+        return u == t[i];
+    });
+}
+
+function prop_equals(u, v) {
+    if (u === v) return true;
+    if (u == null) return v == null;
+    return u instanceof AST_Node && v instanceof AST_Node && u.equals(v);
+}
+
 /* -----[ statements ]----- */
 
 var AST_Statement = DEFNODE("Statement", null, {
@@ -242,6 +266,7 @@ var AST_Statement = DEFNODE("Statement", null, {
 
 var AST_Debugger = DEFNODE("Debugger", null, {
     $documentation: "Represents a debugger statement",
+    _equals: return_true,
 }, AST_Statement);
 
 var AST_Directive = DEFNODE("Directive", "quote value", {
@@ -250,6 +275,9 @@ var AST_Directive = DEFNODE("Directive", "quote value", {
         quote: "[string?] the original quote character",
         value: "[string] The value of this directive as a plain string (it's not an AST_String!)",
     },
+    _equals: function(node) {
+        return this.value == node.value;
+    },
     _validate: function() {
         if (this.quote != null) {
             if (typeof this.quote != "string") throw new Error("quote must be string");
@@ -260,7 +288,8 @@ var AST_Directive = DEFNODE("Directive", "quote value", {
 }, AST_Statement);
 
 var AST_EmptyStatement = DEFNODE("EmptyStatement", null, {
-    $documentation: "The empty statement (empty block or simply a semicolon)"
+    $documentation: "The empty statement (empty block or simply a semicolon)",
+    _equals: return_true,
 }, AST_Statement);
 
 function is_statement(node) {
@@ -291,6 +320,9 @@ var AST_SimpleStatement = DEFNODE("SimpleStatement", "body", {
     $propdoc: {
         body: "[AST_Node] an expression node (should not be instanceof AST_Statement)",
     },
+    _equals: function(node) {
+        return this.body.equals(node.body);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -342,6 +374,9 @@ var AST_Block = DEFNODE("Block", "body", {
     $propdoc: {
         body: "[AST_Statement*] an array of statements"
     },
+    _equals: function(node) {
+        return all_equals(this.body, node.body);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -376,6 +411,10 @@ var AST_LabeledStatement = DEFNODE("LabeledStatement", "label", {
     $propdoc: {
         label: "[AST_Label] a label definition"
     },
+    _equals: function(node) {
+        return this.label.equals(node.label)
+            && this.body.equals(node.body);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -417,6 +456,10 @@ var AST_DWLoop = DEFNODE("DWLoop", "condition", {
     $propdoc: {
         condition: "[AST_Node] the loop condition.  Should not be instanceof AST_Statement"
     },
+    _equals: function(node) {
+        return this.body.equals(node.body)
+            && this.condition.equals(node.condition);
+    },
     _validate: function() {
         if (this.TYPE == "DWLoop") throw new Error("should not instantiate AST_DWLoop");
         must_be_expression(this, "condition");
@@ -431,7 +474,7 @@ var AST_Do = DEFNODE("Do", null, {
             node.body.walk(visitor);
             node.condition.walk(visitor);
         });
-    }
+    },
 }, AST_DWLoop);
 
 var AST_While = DEFNODE("While", null, {
@@ -442,7 +485,7 @@ var AST_While = DEFNODE("While", null, {
             node.condition.walk(visitor);
             node.body.walk(visitor);
         });
-    }
+    },
 }, AST_DWLoop);
 
 var AST_For = DEFNODE("For", "init condition step", {
@@ -452,6 +495,12 @@ var AST_For = DEFNODE("For", "init condition step", {
         condition: "[AST_Node?] the `for` termination clause, or null if empty",
         step: "[AST_Node?] the `for` update clause, or null if empty"
     },
+    _equals: function(node) {
+        return prop_equals(this.init, node.init)
+            && prop_equals(this.condition, node.condition)
+            && prop_equals(this.step, node.step)
+            && this.body.equals(node.body);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -479,6 +528,11 @@ var AST_ForEnumeration = DEFNODE("ForEnumeration", "init object", {
         init: "[AST_Node] the assignment target during iteration",
         object: "[AST_Node] the object to iterate over"
     },
+    _equals: function(node) {
+        return this.init.equals(node.init)
+            && this.object.equals(node.object)
+            && this.body.equals(node.body);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -519,6 +573,10 @@ var AST_With = DEFNODE("With", "expression", {
     $propdoc: {
         expression: "[AST_Node] the `with` expression"
     },
+    _equals: function(node) {
+        return this.expression.equals(node.expression)
+            && this.body.equals(node.body);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -534,7 +592,7 @@ var AST_With = DEFNODE("With", "expression", {
 /* -----[ scope and functions ]----- */
 
 var AST_Scope = DEFNODE("Scope", "fn_defs may_call_this uses_eval uses_with", {
-    $documentation: "Base class for all statements introducing a lexical scope",
+    $documentation: "Base class for all statements introducing a lambda scope",
     $propdoc: {
         uses_eval: "[boolean/S] tells whether this scope contains a direct call to the global `eval`",
         uses_with: "[boolean/S] tells whether this scope uses the `with` statement",
@@ -592,6 +650,10 @@ var AST_Toplevel = DEFNODE("Toplevel", "globals", {
     }
 }, AST_Scope);
 
+var AST_ClassInitBlock = DEFNODE("ClassInitBlock", null, {
+    $documentation: "Value for `class` static initialization blocks",
+}, AST_Scope);
+
 var AST_Lambda = DEFNODE("Lambda", "argnames length_read rest safe_ids uses_arguments", {
     $documentation: "Base class for functions",
     $propdoc: {
@@ -617,6 +679,13 @@ var AST_Lambda = DEFNODE("Lambda", "argnames length_read rest safe_ids uses_argu
         });
         if (this.rest) this.rest.walk(tw);
     },
+    _equals: function(node) {
+        return prop_equals(this.rest, node.rest)
+            && prop_equals(this.name, node.name)
+            && prop_equals(this.value, node.value)
+            && all_equals(this.argnames, node.argnames)
+            && all_equals(this.body, node.body);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -827,6 +896,14 @@ var AST_Class = DEFNODE("Class", "extends name properties", {
         extends: "[AST_Node?] the super class, or null if not specified",
         properties: "[AST_ClassProperty*] array of class properties",
     },
+    _equals: function(node) {
+        return prop_equals(this.name, node.name)
+            && prop_equals(this.extends, node.extends)
+            && all_equals(this.properties, node.properties);
+    },
+    resolve: function(def_class) {
+        return def_class ? this : this.parent_scope.resolve();
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -871,11 +948,17 @@ var AST_ClassExpression = DEFNODE("ClassExpression", null, {
 var AST_ClassProperty = DEFNODE("ClassProperty", "key private static value", {
     $documentation: "Base class for `class` properties",
     $propdoc: {
-        key: "[string|AST_Node] property name (AST_Node for computed property)",
+        key: "[string|AST_Node?] property name (AST_Node for computed property, null for initialization block)",
         private: "[boolean] whether this is a private property",
         static: "[boolean] whether this is a static property",
         value: "[AST_Node?] property value (AST_Accessor for getters/setters, AST_LambdaExpression for methods, null if not specified for fields)",
     },
+    _equals: function(node) {
+        return !this.private == !node.private
+            && !this.static == !node.static
+            && prop_equals(this.key, node.key)
+            && prop_equals(this.value, node.value);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -885,7 +968,9 @@ var AST_ClassProperty = DEFNODE("ClassProperty", "key private static value", {
     },
     _validate: function() {
         if (this.TYPE == "ClassProperty") throw new Error("should not instantiate AST_ClassProperty");
-        if (typeof this.key != "string") {
+        if (this instanceof AST_ClassInit) {
+            if (this.key != null) throw new Error("key must be null");
+        } else if (typeof this.key != "string") {
             if (!(this.key instanceof AST_Node)) throw new Error("key must be string or AST_Node");
             must_be_expression(this, "key");
         }
@@ -925,6 +1010,17 @@ var AST_ClassMethod = DEFNODE("ClassMethod", null, {
     },
 }, AST_ClassProperty);
 
+var AST_ClassInit = DEFNODE("ClassInit", null, {
+    $documentation: "A `class` static initialization block",
+    _validate: function() {
+        if (!this.static) throw new Error("static must be true");
+        if (!(this.value instanceof AST_ClassInitBlock)) throw new Error("value must be AST_ClassInitBlock");
+    },
+    initialize: function() {
+        this.static = true;
+    },
+}, AST_ClassProperty);
+
 /* -----[ JUMPS ]----- */
 
 var AST_Jump = DEFNODE("Jump", null, {
@@ -939,6 +1035,9 @@ var AST_Exit = DEFNODE("Exit", "value", {
     $propdoc: {
         value: "[AST_Node?] the value returned or thrown by this statement; could be null for AST_Return"
     },
+    _equals: function(node) {
+        return prop_equals(this.value, node.value);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -969,6 +1068,9 @@ var AST_LoopControl = DEFNODE("LoopControl", "label", {
     $propdoc: {
         label: "[AST_LabelRef?] the label, or null if none",
     },
+    _equals: function(node) {
+        return prop_equals(this.label, node.label);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -999,6 +1101,11 @@ var AST_If = DEFNODE("If", "condition alternative", {
         condition: "[AST_Node] the `if` condition",
         alternative: "[AST_Statement?] the `else` part, or null if not present"
     },
+    _equals: function(node) {
+        return this.body.equals(node.body)
+            && this.condition.equals(node.condition)
+            && prop_equals(this.alternative, node.alternative);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1022,6 +1129,10 @@ var AST_Switch = DEFNODE("Switch", "expression", {
     $propdoc: {
         expression: "[AST_Node] the `switch` “discriminant”"
     },
+    _equals: function(node) {
+        return this.expression.equals(node.expression)
+            && all_equals(this.body, node.body);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1053,6 +1164,10 @@ var AST_Case = DEFNODE("Case", "expression", {
     $propdoc: {
         expression: "[AST_Node] the `case` expression"
     },
+    _equals: function(node) {
+        return this.expression.equals(node.expression)
+            && all_equals(this.body, node.body);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1073,6 +1188,11 @@ var AST_Try = DEFNODE("Try", "bcatch bfinally", {
         bcatch: "[AST_Catch?] the catch block, or null if not present",
         bfinally: "[AST_Finally?] the finally block, or null if not present"
     },
+    _equals: function(node) {
+        return all_equals(this.body, node.body)
+            && prop_equals(this.bcatch, node.bcatch)
+            && prop_equals(this.bfinally, node.bfinally);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1096,6 +1216,10 @@ var AST_Catch = DEFNODE("Catch", "argname", {
     $propdoc: {
         argname: "[(AST_Destructured|AST_SymbolCatch)?] symbol for the exception, or null if not present",
     },
+    _equals: function(node) {
+        return prop_equals(this.argname, node.argname)
+            && all_equals(this.body, node.body);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1121,6 +1245,9 @@ var AST_Definitions = DEFNODE("Definitions", "definitions", {
     $propdoc: {
         definitions: "[AST_VarDef*] array of variable definitions"
     },
+    _equals: function(node) {
+        return all_equals(this.definitions, node.definitions);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1177,6 +1304,10 @@ var AST_VarDef = DEFNODE("VarDef", "name value", {
         name: "[AST_Destructured|AST_SymbolVar] name of the variable",
         value: "[AST_Node?] initializer, or null of there's no initializer",
     },
+    _equals: function(node) {
+        return this.name.equals(node.name)
+            && prop_equals(this.value, node.value);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1196,6 +1327,9 @@ var AST_ExportDeclaration = DEFNODE("ExportDeclaration", "body", {
     $propdoc: {
         body: "[AST_DefClass|AST_Definitions|AST_LambdaDefinition] the statement to export",
     },
+    _equals: function(node) {
+        return this.body.equals(node.body);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1216,6 +1350,9 @@ var AST_ExportDefault = DEFNODE("ExportDefault", "body", {
     $propdoc: {
         body: "[AST_Node] the default export",
     },
+    _equals: function(node) {
+        return this.body.equals(node.body);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1229,29 +1366,29 @@ var AST_ExportDefault = DEFNODE("ExportDefault", "body", {
     },
 }, AST_Statement);
 
-var AST_ExportForeign = DEFNODE("ExportForeign", "aliases keys path quote", {
+var AST_ExportForeign = DEFNODE("ExportForeign", "aliases keys path", {
     $documentation: "An `export ... from '...'` statement",
     $propdoc: {
-        aliases: "[string*] array of aliases to export",
-        keys: "[string*] array of keys to import",
-        path: "[string] the path to import module",
-        quote: "[string?] the original quote character",
+        aliases: "[AST_String*] array of aliases to export",
+        keys: "[AST_String*] array of keys to import",
+        path: "[AST_String] the path to import module",
+    },
+    _equals: function(node) {
+        return this.path.equals(node.path)
+            && all_equals(this.aliases, node.aliases)
+            && all_equals(this.keys, node.keys);
     },
     _validate: function() {
         if (this.aliases.length != this.keys.length) {
             throw new Error("aliases:key length mismatch: " + this.aliases.length + " != " + this.keys.length);
         }
         this.aliases.forEach(function(name) {
-            if (typeof name != "string") throw new Error("aliases must contain string");
+            if (!(name instanceof AST_String)) throw new Error("aliases must contain AST_String");
         });
         this.keys.forEach(function(name) {
-            if (typeof name != "string") throw new Error("keys must contain string");
+            if (!(name instanceof AST_String)) throw new Error("keys must contain AST_String");
         });
-        if (typeof this.path != "string") throw new Error("path must be string");
-        if (this.quote != null) {
-            if (typeof this.quote != "string") throw new Error("quote must be string");
-            if (!/^["']$/.test(this.quote)) throw new Error("invalid quote: " + this.quote);
-        }
+        if (!(this.path instanceof AST_String)) throw new Error("path must be AST_String");
     },
 }, AST_Statement);
 
@@ -1260,6 +1397,9 @@ var AST_ExportReferences = DEFNODE("ExportReferences", "properties", {
     $propdoc: {
         properties: "[AST_SymbolExport*] array of aliases to export",
     },
+    _equals: function(node) {
+        return all_equals(this.properties, node.properties);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1275,14 +1415,20 @@ var AST_ExportReferences = DEFNODE("ExportReferences", "properties", {
     },
 }, AST_Statement);
 
-var AST_Import = DEFNODE("Import", "all default path properties quote", {
+var AST_Import = DEFNODE("Import", "all default path properties", {
     $documentation: "An `import` statement",
     $propdoc: {
         all: "[AST_SymbolImport?] the imported namespace, or null if not specified",
         default: "[AST_SymbolImport?] the alias for default `export`, or null if not specified",
-        path: "[string] the path to import module",
+        path: "[AST_String] the path to import module",
         properties: "[(AST_SymbolImport*)?] array of aliases, or null if not specified",
-        quote: "[string?] the original quote character",
+    },
+    _equals: function(node) {
+        return this.path.equals(node.path)
+            && prop_equals(this.all, node.all)
+            && prop_equals(this.default, node.default)
+            && !this.properties == !node.properties
+            && (!this.properties || all_equals(this.properties, node.properties));
     },
     walk: function(visitor) {
         var node = this;
@@ -1301,16 +1447,12 @@ var AST_Import = DEFNODE("Import", "all default path properties quote", {
         }
         if (this.default != null) {
             if (!(this.default instanceof AST_SymbolImport)) throw new Error("default must be AST_SymbolImport");
-            if (this.default.key !== "") throw new Error("invalid default key: " + this.default.key);
+            if (this.default.key.value !== "") throw new Error("invalid default key: " + this.default.key.value);
         }
-        if (typeof this.path != "string") throw new Error("path must be string");
+        if (!(this.path instanceof AST_String)) throw new Error("path must be AST_String");
         if (this.properties != null) this.properties.forEach(function(node) {
             if (!(node instanceof AST_SymbolImport)) throw new Error("properties must contain AST_SymbolImport");
         });
-        if (this.quote != null) {
-            if (typeof this.quote != "string") throw new Error("quote must be string");
-            if (!/^["']$/.test(this.quote)) throw new Error("invalid quote: " + this.quote);
-        }
     },
 }, AST_Statement);
 
@@ -1320,6 +1462,10 @@ var AST_DefaultValue = DEFNODE("DefaultValue", "name value", {
         name: "[AST_Destructured|AST_SymbolDeclaration] name of the variable",
         value: "[AST_Node] value to assign if variable is `undefined`",
     },
+    _equals: function(node) {
+        return this.name.equals(node.name)
+            && this.value.equals(node.value);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1347,6 +1493,11 @@ var AST_Call = DEFNODE("Call", "args expression optional pure terminal", {
         pure: "[boolean/S] marker for side-effect-free call expression",
         terminal: "[boolean] whether the chain has ended",
     },
+    _equals: function(node) {
+        return !this.optional == !node.optional
+            && this.expression.equals(node.expression)
+            && all_equals(this.args, node.args);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1375,6 +1526,9 @@ var AST_Sequence = DEFNODE("Sequence", "expressions", {
     $propdoc: {
         expressions: "[AST_Node*] array of expressions (at least two)"
     },
+    _equals: function(node) {
+        return all_equals(this.expressions, node.expressions);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1402,6 +1556,11 @@ var AST_PropAccess = DEFNODE("PropAccess", "expression optional property termina
         property: "[AST_Node|string] the property to access.  For AST_Dot this is always a plain string, while for AST_Sub it's an arbitrary AST_Node",
         terminal: "[boolean] whether the chain has ended",
     },
+    _equals: function(node) {
+        return !this.optional == !node.optional
+            && prop_equals(this.property, node.property)
+            && this.expression.equals(node.expression);
+    },
     get_property: function() {
         var p = this.property;
         if (p instanceof AST_Constant) return p.value;
@@ -1446,6 +1605,9 @@ var AST_Spread = DEFNODE("Spread", "expression", {
     $propdoc: {
         expression: "[AST_Node] expression to be expanded",
     },
+    _equals: function(node) {
+        return this.expression.equals(node.expression);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1463,6 +1625,10 @@ var AST_Unary = DEFNODE("Unary", "operator expression", {
         operator: "[string] the operator",
         expression: "[AST_Node] expression that this unary operator applies to"
     },
+    _equals: function(node) {
+        return this.operator == node.operator
+            && this.expression.equals(node.expression);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1491,6 +1657,11 @@ var AST_Binary = DEFNODE("Binary", "operator left right", {
         operator: "[string] the operator",
         right: "[AST_Node] right-hand side expression"
     },
+    _equals: function(node) {
+        return this.operator == node.operator
+            && this.left.equals(node.left)
+            && this.right.equals(node.right);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1512,6 +1683,11 @@ var AST_Conditional = DEFNODE("Conditional", "condition consequent alternative",
         consequent: "[AST_Node]",
         alternative: "[AST_Node]"
     },
+    _equals: function(node) {
+        return this.condition.equals(node.condition)
+            && this.consequent.equals(node.consequent)
+            && this.alternative.equals(node.alternative);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1553,6 +1729,9 @@ var AST_Await = DEFNODE("Await", "expression", {
     $propdoc: {
         expression: "[AST_Node] expression with Promise to resolve on",
     },
+    _equals: function(node) {
+        return this.expression.equals(node.expression);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1570,6 +1749,10 @@ var AST_Yield = DEFNODE("Yield", "expression nested", {
         expression: "[AST_Node?] return value for iterator, or null if undefined",
         nested: "[boolean] whether to iterate over expression as generator",
     },
+    _equals: function(node) {
+        return !this.nested == !node.nested
+            && prop_equals(this.expression, node.expression);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1592,6 +1775,9 @@ var AST_Array = DEFNODE("Array", "elements", {
     $propdoc: {
         elements: "[AST_Node*] array of elements"
     },
+    _equals: function(node) {
+        return all_equals(this.elements, node.elements);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1634,6 +1820,10 @@ var AST_DestructuredArray = DEFNODE("DestructuredArray", "elements", {
     $propdoc: {
         elements: "[(AST_DefaultValue|AST_Destructured|AST_SymbolDeclaration|AST_SymbolRef)*] array of elements",
     },
+    _equals: function(node) {
+        return prop_equals(this.rest, node.rest)
+            && all_equals(this.elements, node.elements);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1651,6 +1841,10 @@ var AST_DestructuredKeyVal = DEFNODE("DestructuredKeyVal", "key value", {
         key: "[string|AST_Node] property name.  For computed property this is an AST_Node.",
         value: "[AST_DefaultValue|AST_Destructured|AST_SymbolDeclaration|AST_SymbolRef] property value",
     },
+    _equals: function(node) {
+        return prop_equals(this.key, node.key)
+            && this.value.equals(node.value);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1672,6 +1866,10 @@ var AST_DestructuredObject = DEFNODE("DestructuredObject", "properties", {
     $propdoc: {
         properties: "[AST_DestructuredKeyVal*] array of properties",
     },
+    _equals: function(node) {
+        return prop_equals(this.rest, node.rest)
+            && all_equals(this.properties, node.properties);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1693,6 +1891,9 @@ var AST_Object = DEFNODE("Object", "properties", {
     $propdoc: {
         properties: "[(AST_ObjectProperty|AST_Spread)*] array of properties"
     },
+    _equals: function(node) {
+        return all_equals(this.properties, node.properties);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1716,6 +1917,10 @@ var AST_ObjectProperty = DEFNODE("ObjectProperty", "key value", {
         key: "[string|AST_Node] property name.  For computed property this is an AST_Node.",
         value: "[AST_Node] property value.  For getters and setters this is an AST_Accessor.",
     },
+    _equals: function(node) {
+        return prop_equals(this.key, node.key)
+            && this.value.equals(node.value);
+    },
     walk: function(visitor) {
         var node = this;
         visitor.visit(node, function() {
@@ -1770,6 +1975,9 @@ var AST_Symbol = DEFNODE("Symbol", "scope name thedef", {
         scope: "[AST_Scope/S] the current scope (not necessarily the definition scope)",
         thedef: "[SymbolDef/S] the definition of this symbol"
     },
+    _equals: function(node) {
+        return this.thedef ? this.thedef === node.thedef : this.name == node.name;
+    },
     _validate: function() {
         if (this.TYPE == "Symbol") throw new Error("should not instantiate AST_Symbol");
         if (typeof this.name != "string") throw new Error("name must be string");
@@ -1787,10 +1995,14 @@ var AST_SymbolConst = DEFNODE("SymbolConst", null, {
 var AST_SymbolImport = DEFNODE("SymbolImport", "key", {
     $documentation: "Symbol defined by an `import` statement",
     $propdoc: {
-        key: "[string] the original `export` name",
+        key: "[AST_String] the original `export` name",
+    },
+    _equals: function(node) {
+        return this.name == node.name
+            && this.key.equals(node.key);
     },
     _validate: function() {
-        if (typeof this.key != "string") throw new Error("key must be string");
+        if (!(this.key instanceof AST_String)) throw new Error("key must be AST_String");
     },
 }, AST_SymbolConst);
 
@@ -1834,7 +2046,7 @@ var AST_Label = DEFNODE("Label", "references", {
     initialize: function() {
         this.references = [];
         this.thedef = this;
-    }
+    },
 }, AST_Symbol);
 
 var AST_SymbolRef = DEFNODE("SymbolRef", "fixed in_arg redef", {
@@ -1844,10 +2056,14 @@ var AST_SymbolRef = DEFNODE("SymbolRef", "fixed in_arg redef", {
 var AST_SymbolExport = DEFNODE("SymbolExport", "alias", {
     $documentation: "Reference in an `export` statement",
     $propdoc: {
-        alias: "[string] the `export` alias",
+        alias: "[AST_String] the `export` alias",
+    },
+    _equals: function(node) {
+        return this.name == node.name
+            && this.alias.equals(node.alias);
     },
     _validate: function() {
-        if (typeof this.alias != "string") throw new Error("alias must be string");
+        if (!(this.alias instanceof AST_String)) throw new Error("alias must be AST_String");
     },
 }, AST_SymbolRef);
 
@@ -1857,6 +2073,7 @@ var AST_LabelRef = DEFNODE("LabelRef", null, {
 
 var AST_ObjectIdentity = DEFNODE("ObjectIdentity", null, {
     $documentation: "Base class for `super` & `this`",
+    _equals: return_true,
     _validate: function() {
         if (this.TYPE == "ObjectIdentity") throw new Error("should not instantiate AST_ObjectIdentity");
     },
@@ -1891,7 +2108,12 @@ var AST_Template = DEFNODE("Template", "expressions strings tag", {
     $propdoc: {
         expressions: "[AST_Node*] the placeholder expressions",
         strings: "[string*] the raw text segments",
-        tag: "[AST_Node] tag function, or null if absent",
+        tag: "[AST_Node?] tag function, or null if absent",
+    },
+    _equals: function(node) {
+        return prop_equals(this.tag, node.tag)
+            && list_equals(this.strings, node.strings)
+            && all_equals(this.expressions, node.expressions);
     },
     walk: function(visitor) {
         var node = this;
@@ -1916,6 +2138,9 @@ var AST_Template = DEFNODE("Template", "expressions strings tag", {
 
 var AST_Constant = DEFNODE("Constant", null, {
     $documentation: "Base class for all constants",
+    _equals: function(node) {
+        return this.value === node.value;
+    },
     _validate: function() {
         if (this.TYPE == "Constant") throw new Error("should not instantiate AST_Constant");
     },
@@ -1964,6 +2189,9 @@ var AST_RegExp = DEFNODE("RegExp", "value", {
     $propdoc: {
         value: "[RegExp] the actual regexp"
     },
+    _equals: function(node) {
+        return "" + this.value == "" + node.value;
+    },
     _validate: function() {
         if (!(this.value instanceof RegExp)) throw new Error("value must be RegExp");
     },
@@ -1971,6 +2199,7 @@ var AST_RegExp = DEFNODE("RegExp", "value", {
 
 var AST_Atom = DEFNODE("Atom", null, {
     $documentation: "Base class for atoms",
+    _equals: return_true,
     _validate: function() {
         if (this.TYPE == "Atom") throw new Error("should not instantiate AST_Atom");
     },
@@ -2036,16 +2265,21 @@ TreeWalker.prototype = {
         return this.stack[this.stack.length - 2 - (n || 0)];
     },
     push: function(node) {
-        if (node instanceof AST_Lambda) {
+        var value;
+        if (node instanceof AST_Class) {
+            this.directives = Object.create(this.directives);
+            value = "use strict";
+        } else if (node instanceof AST_Directive) {
+            value = node.value;
+        } else if (node instanceof AST_Lambda) {
             this.directives = Object.create(this.directives);
-        } else if (node instanceof AST_Directive && !this.directives[node.value]) {
-            this.directives[node.value] = node;
         }
+        if (value && !this.directives[value]) this.directives[value] = node;
         this.stack.push(node);
     },
     pop: function() {
         var node = this.stack.pop();
-        if (node instanceof AST_Lambda) {
+        if (node instanceof AST_Class || node instanceof AST_Lambda) {
             this.directives = Object.getPrototypeOf(this.directives);
         }
     },
@@ -2085,33 +2319,40 @@ TreeWalker.prototype = {
         }
     },
     in_boolean_context: function() {
-        var self = this.self();
-        for (var i = 0, p; p = this.parent(i); i++) {
-            if (p instanceof AST_Conditional && p.condition === self
-                || p instanceof AST_DWLoop && p.condition === self
-                || p instanceof AST_For && p.condition === self
-                || p instanceof AST_If && p.condition === self
-                || p instanceof AST_Return && p.in_bool
-                || p instanceof AST_Sequence && p.tail_node() !== self
-                || p instanceof AST_SimpleStatement
-                || p instanceof AST_UnaryPrefix && p.operator == "!" && p.expression === self) {
-                return true;
-            }
-            if (p instanceof AST_Binary && (p.operator == "&&" || p.operator == "||")
-                || p instanceof AST_Conditional
-                || p.tail_node() === self) {
-                self = p;
-            } else if (p instanceof AST_Return) {
-                for (var call, fn = p; call = this.parent(++i); fn = call) {
-                    if (call.TYPE == "Call") {
-                        if (!(fn instanceof AST_Lambda) || fn.name) return false;
-                    } else if (fn instanceof AST_Lambda) {
-                        return false;
-                    }
-                }
-            } else {
+        for (var drop = true, level = 0, parent, self = this.self(); parent = this.parent(level++); self = parent) {
+            if (parent instanceof AST_Binary) switch (parent.operator) {
+              case "&&":
+              case "||":
+                if (parent.left === self) drop = false;
+                continue;
+              default:
                 return false;
             }
+            if (parent instanceof AST_Conditional) {
+                if (parent.condition === self) return true;
+                continue;
+            }
+            if (parent instanceof AST_DWLoop) return parent.condition === self;
+            if (parent instanceof AST_For) return parent.condition === self;
+            if (parent instanceof AST_If) return parent.condition === self;
+            if (parent instanceof AST_Return) {
+                if (parent.in_bool) return true;
+                while (parent = this.parent(level++)) {
+                    if (parent instanceof AST_Lambda) {
+                        if (parent.name) return false;
+                        parent = this.parent(level++);
+                        if (parent.TYPE != "Call") return false;
+                        break;
+                    }
+                }
+            }
+            if (parent instanceof AST_Sequence) {
+                if (parent.tail_node() === self) continue;
+                return drop ? "d" : true;
+            }
+            if (parent instanceof AST_SimpleStatement) return drop ? "d" : true;
+            if (parent instanceof AST_UnaryPrefix) return parent.operator == "!";
+            return false;
         }
     }
 };
